import torch
import os
import pickle
import click
import wandb
from nanoppo.continuous_action_ppo import PPOAgent
from nanoppo.normalizer import Normalizer
from nanoppo.ppo_utils import compute_gae
from nanoppo.environment_manager import EnvironmentManager
from nanoppo.ppo_utils import get_grad_norm

# Memory for PPO
class PPOMemory:
    def __init__(self, device):
        self.states = []
        self.actions = []
        self.logprobs = []
        self.next_states = []
        self.rewards = []
        self.is_terminals = []
        self.device = device

    def clear(self):
        del self.states[:]
        del self.actions[:]
        del self.logprobs[:]
        del self.next_states[:]
        del self.rewards[:]
        del self.is_terminals[:]

    def append(self, state, action, logprob, next_state, reward, is_terminal):
        assert (
            len(self.states)
            == len(self.actions)
            == len(self.logprobs)
            == len(self.next_states)
            == len(self.rewards)
            == len(self.is_terminals)
        )
        self.states.append(state)
        self.actions.append(action)
        self.logprobs.append(logprob)
        self.next_states.append(next_state)
        self.rewards.append(
            torch.tensor(reward, dtype=torch.float32).to(self.device)
        )  # Convert to tensor here
        self.is_terminals.append(
            torch.tensor(is_terminal, dtype=torch.float32).to(self.device)
        )  # Convert to tensor here

    def get(self):
        return (
            torch.stack(self.states),
            torch.stack(self.actions),
            torch.stack(self.logprobs),
            torch.stack(self.next_states),
            torch.stack(self.rewards),
            torch.stack(self.is_terminals),
        )


def train_agent(
    env_name,
    env_config = None,
    max_episodes=500,
    stop_reward = None,
    policy_class = None,
    policy_lr=0.0005,
    value_lr=0.0005,
    betas=(0.9, 0.999),
    n_latent_var=128,
    gamma=0.99,
    tau=0.95,
    K_epochs=4,
    eps_clip=0.2,
    vl_coef=0.5,
    el_coef=0.001,
    max_timesteps=2000,
    update_timestep=200,
    checkpoint_dir="checkpoints",
    checkpoint_interval=-1,
    log_interval=-1,
    lr_scheduler=None,
    wandb_log=False,
    device=None,
    debug=False,
):
    if device is None:
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    # Setting up the environment and the agent
    env = EnvironmentManager(env_name, env_config).setup_env()
    state_dim = env.observation_space.shape[-1]
    action_dim = env.action_space.shape[0]
    print("state_dim", state_dim)
    print("action_dim", action_dim)

    checkpoint_path = os.path.join(checkpoint_dir, env_name)
    os.makedirs(checkpoint_path, exist_ok=True)
    print("env", env_name)
    model_file = f"{checkpoint_path}/models.pth"
    metrics_file = f"{checkpoint_path}/metrics.pkl"
    if wandb_log:
        wandb.init(
            project="nanoPPO",
            name=env_name,
            config={
                "policy_lr": policy_lr,
                "value_lr": value_lr,
                "betas": betas,
                "gamma": gamma,
                "tau": tau,
                "K_epochs": K_epochs,
                "eps_clip": eps_clip,
            },
        )

    # Initialize a normalizer with the dimensionality of the state
    state_normalizer = Normalizer(dim=env.observation_space.shape)
    ppo = PPOAgent(
        state_dim,
        action_dim,
        n_latent_var,
        policy_class,
        policy_lr,
        value_lr,
        betas,
        gamma,
        K_epochs,
        eps_clip,
        state_normalizer,
        action_low=env.action_space.low,
        action_high=env.action_space.high,
        vl_coef=vl_coef,
        el_coef=el_coef,
        lr_scheduler=lr_scheduler,
        device=device,
        wandb_log=wandb_log,
        debug = debug
    )
    print(policy_lr, value_lr, betas)
    print('ppo use device', ppo.device)

    # Load the best weights
    if os.path.exists(model_file):
        metrics = pickle.load(open(metrics_file, "rb"))
        best_reward = metrics["best_reward"] 
        start_episode = metrics["episode"] + 1
        ppo.load(model_file)
        print("Loaded best weights!", model_file, metrics_file)
        if stop_reward and (best_reward > stop_reward):
            print("Skipping Training: best_reward", best_reward, "> stop_reward", stop_reward)
            return ppo, model_file, metrics_file
    else:
        best_reward = float("-inf")
        start_episode = 1
    print("best_reward", best_reward)
    print("start_episode", start_episode)
    print("log_interval", log_interval)

    ppo_memory = PPOMemory(device= device)

    # Training loop
    time_step = 0
    avg_length_list = []
    cumulative_reward_list = []  # Initialize cumulative reward
    for episode in range(start_episode, max_episodes + start_episode):
        state, info = env.reset()
        state_normalizer.observe(state)
        state = state_normalizer.normalize(state)

        total_reward = 0
        state = torch.FloatTensor(state).to(device)
        for t in range(max_timesteps):
            action, log_prob = ppo.policy.act(state)

            action_np = action.detach().cpu().numpy()
            if len(action_np.shape) > 1:
                # (1, action_dim) -> (action_dim,)
                action_np = action_np.squeeze()
            next_state, reward, done, truncated, _ = env.step(action_np)
            state_normalizer.observe(next_state)
            next_state = state_normalizer.normalize(next_state)

            total_reward += reward
            next_state = torch.FloatTensor(next_state).to(device)
            ppo_memory.append(state, action, log_prob, next_state, reward, done)

            state = next_state
            time_step += 1

            # update if it's time
            if time_step % update_timestep == 0:
                try:
                    # Get state values for all states
                    next_value = ppo.policy.get_value(next_state).detach().item()
                    values = [
                        ppo.policy.get_value(state).item()
                        for state in ppo_memory.states
                    ]
                    masks = [1 - terminal.item() for terminal in ppo_memory.is_terminals]
                    returns = compute_gae(
                        next_value, ppo_memory.rewards, masks, values, gamma=gamma, tau=tau
                    )
                    torch_returns = torch.tensor(returns, dtype=torch.float32).to(device)
    
                    (
                        states,
                        actions,
                        log_probs,
                        next_states,
                        rewards,
                        dones,
                    ) = ppo_memory.get()
                    ppo.update(
                        states,
                        actions,
                        returns=torch_returns,
                        next_states=next_states,
                        dones=dones,
                    )
                    ppo_memory.clear()
                    time_step = 0
                except Exception as e:
                    print("ppo.update error")
                    print(e)
                    breakpoint()
                    raise e
            if done or truncated:
                break
        avg_length_list.append(t + 1)

        cumulative_reward_list.append(total_reward)

        num_cumulative_rewards = len(cumulative_reward_list)
        avg_reward = float(sum(cumulative_reward_list[-30:]) / 30)
        avg_length = int(sum(avg_length_list) / len(avg_length_list))
        action_mu_grad_norm = get_grad_norm(ppo.policy.action_mu.parameters())
        action_log_std_grad_norm = get_grad_norm(ppo.policy.action_log_std.parameters())
        value_grad_norm = get_grad_norm(ppo.policy.value_layer.parameters())
        # Logging
        if log_interval > 0 and (episode % log_interval == 0):
            sample_length = len(avg_length_list)
            avg_length = int(sum(avg_length_list) / sample_length)
            print(
                (
                    "Episode {} \t samples:{} avg steps: {} \t avg reward: {:.3f} \t best reward: {:.3f} \t"
                    "action_mu_grad_norm: {:.2f} \t action_log_std_grad_norm: {:.2f} \t value_grad_norm: {:.2f} \t"
                    "num cumulative rewards: {}"
                ).format(
                    episode,
                    sample_length,
                    avg_length,
                    avg_reward,
                    best_reward,
                    action_mu_grad_norm,
                    action_log_std_grad_norm,
                    value_grad_norm,
                    num_cumulative_rewards
                )
            )

        if (checkpoint_interval > 0 and (avg_reward > best_reward) and (num_cumulative_rewards > 30)):
            print("avg_reward", avg_reward, "> best_reward", best_reward)
            best_reward = avg_reward
            metrics = {"train_reward": avg_reward, "best_reward": best_reward, "episode": episode, "stop_reward":stop_reward}
            pickle.dump(metrics, open(metrics_file, "wb"))
            ppo.save(model_file)
            print("Saved best weights!", best_reward, model_file, metrics_file)
        
        if stop_reward and (avg_reward > stop_reward) and (num_cumulative_rewards > 30):
            print("avg_reward", avg_reward, "> stop_reward", stop_reward)
            best_reward = avg_reward
            metrics = {"train_reward":avg_reward, "best_reward": best_reward, "episode": episode, "stop_reward":stop_reward}
            pickle.dump(metrics, open(metrics_file, "wb"))
            ppo.save(model_file)
            print("Saved best weights!", best_reward, model_file, metrics_file)
            break

        if wandb_log:
            wandb.log(
                {
                    "avg_reward": avg_reward,
                    "best_reward": best_reward,
                    "avg_length": avg_length,
                    "action_mu_grad_norm": action_mu_grad_norm,
                    "action_log_std_grad_norm": action_log_std_grad_norm,
                    "value_grad_norm": value_grad_norm,
                }
            )
    if wandb_log:
        wandb.finish()
    return ppo, model_file, metrics_file


@click.command()
@click.option(
    "--env_name",
    default="PointMass2D-v0",
    type=click.Choice(
        ["PointMass1D-v0", "PointMass2D-v0", "Pendulum-v1", "MountainCarContinuous-v0"]
    ),
)
@click.option("--max_episodes", default=100, help="Number of training episodes.")
@click.option("--policy_lr", default=0.0005, help="Learning rate for policy network.")
@click.option("--value_lr", default=0.0005, help="Learning rate for value network.")
@click.option("--vl_coef", default=0.5, help="Value function coefficient.")
@click.option("--checkpoint_dir", default="checkpoints", help="Path to checkpoint.")
@click.option("--checkpoint_interval", default=100, help="Checkpoint interval.")
@click.option("--log_interval", default=10, help="Logging interval.")
@click.option(
    "--wandb_log", is_flag=True, default=False, help="Flag to log results to wandb."
)
def cli(
    env_name,
    max_episodes,
    policy_lr,
    value_lr,
    vl_coef,
    checkpoint_dir,
    checkpoint_interval,
    log_interval,
    wandb_log,
):
    ppo, model_file, metrics_file = train_agent(
        env_name=env_name,
        max_episodes=max_episodes,
        policy_lr=policy_lr,
        value_lr=value_lr,
        vl_coef=vl_coef,
        checkpoint_dir=checkpoint_dir,
        checkpoint_interval=checkpoint_interval,
        log_interval=log_interval,
        wandb_log=wandb_log,
        device='cpu'
    )
    # Load the best weights
    ppo.load(model_file)
    print("Loaded best weights from", model_file)
    metrics = pickle.load(open(metrics_file, "rb"))
    print("Loaded metrics from", metrics_file)
    best_reward = metrics["best_reward"]
    episode = metrics["episode"]
    print("best_reward", best_reward, "episode", episode)


if __name__ == "__main__":
    cli()



import torch
import torch.nn as nn
import torch.optim as optim

# Neural Network for Actor-Critic
import torch.nn as nn
import torch.nn.functional as F
from nanoppo.policy.actor_critic import ActorCritic
import wandb


# PPO Agent
class PPOAgent:
    def __init__(
        self,
        state_dim,
        action_dim,
        n_latent_var,
        policy_class,
        policy_lr,
        value_lr,
        betas,
        gamma,
        K_epochs,
        eps_clip,
        state_normalizer,
        action_low,
        action_high,
        vl_coef=0.5,
        el_coef=0.001,
        lr_scheduler=None,  # Add lr_scheduler as an optional argument
        device="cpu",
        wandb_log=False,
        debug = False
    ):
        self.gamma = gamma
        self.eps_clip = eps_clip
        self.K_epochs = K_epochs
        self.state_normalizer = state_normalizer
        self.action_low = action_low
        self.action_high = action_high
        self.vl_coef = vl_coef
        self.el_coef = el_coef
        self.device = device
        self.debug = debug

        # Initialize optimizer with a placeholder
        self.optimizer = None

        action_low_tensor = torch.tensor(action_low, dtype=torch.float32).to(device)
        action_high_tensor = torch.tensor(action_high, dtype=torch.float32).to(device)
        
        if policy_class:
            self.policy = policy_class(
                state_dim, action_dim, n_latent_var, action_low_tensor, action_high_tensor, device=device, debug=debug
            ).float().to(device)
            self.policy_old = policy_class(
                state_dim, action_dim, n_latent_var, action_low_tensor, action_high_tensor, device=device, debug=debug
            ).float().to(device)
            self.policy_old.load_state_dict(self.policy.state_dict())
        else:
            self.policy = ActorCritic(
                state_dim, action_dim, n_latent_var, action_low_tensor, action_high_tensor, debug=debug
            ).float().to(device)
            self.policy_old = ActorCritic(
                state_dim, action_dim, n_latent_var, action_low_tensor, action_high_tensor, debug=debug
            ).float().to(device)
            self.policy_old.load_state_dict(self.policy.state_dict())

        # Separate the parameters of the actor and critic networks
        actor_params = list(self.policy.action_mu.parameters()) + list(
            self.policy.action_log_std.parameters()
        )
        critic_params = list(self.policy.value_layer.parameters())

        # Use the learning rates from the lr_scheduler if provided, or use the original learning rates
        if lr_scheduler is not None:
            actor_lr = lr_scheduler.get_lr_actor(it=0)
            critic_lr = lr_scheduler.get_lr_critic(it=0)
        else:
            actor_lr = policy_lr
            critic_lr = value_lr

        # Create an optimizer with different learning rates for the actor and critic
        self.optimizer = torch.optim.Adam(
            [
                {"params": actor_params, "lr": actor_lr},
                {"params": critic_params, "lr": critic_lr},
            ],
            betas=betas,
        )

        self.mse_loss = nn.MSELoss()
        self.wandb_log = wandb_log

        # Store the lr_scheduler if provided
        self.lr_scheduler = lr_scheduler

        self.iterations = 0

    def check_weights(self, model_part):
        if self.debug:
            for name, param in model_part.named_parameters():
                if torch.isnan(param.data).any():
                    print(f"NaN detected in weights: {name}")
                    # Here you can call your debugging method or do other appropriate handling
                    self.debug_this(model_part)
                    breakpoint()  # or raise an exception, or log the issue, etc.
    
    def check_gradients(self, model_part):
        if self.debug:
            for name, param in model_part.named_parameters():
                if param.grad is not None and torch.isnan(param.grad).any():
                    print(f"NaN detected in gradients: {name}")
                    # Here you can call your debugging method or do other appropriate handling
                    self.debug_this(model_part)
                    breakpoint()  # or raise an exception, or log the issue, etc.
        
    # Define the debug function
    def debug_this(self, model, state = None):
        print("Debugging Information:")
        # Check and print the weights and biases
        print("\nWeights and Biases in each layer:")
        for name, param in model.named_parameters():
            print(name, param.data)
        
        # Pass the input through each layer individually and print the outputs
        if state:
            x = state
            for i, layer in enumerate(model):
                x = layer(x)
                print(f"\nOutput after layer {i} ({layer.__class__.__name__}):")
                print(x)
                # If you find NaN at this stage, you can break the loop for efficiency
                if torch.isnan(x).any():
                    print(f"Stopping early due to NaN at layer {i}")
                    break

    def update(self, states, actions, returns, next_states, dones):
        # Before backpropagation: Check for NaN
        self.check_weights(self.policy.action_mu)

        for _ in range(self.K_epochs):
            if self.debug:
                # Check the actions for invalid values
                if torch.isnan(actions).any() or torch.isinf(actions).any():
                    print("Invalid values detected in 'actions'")
                    breakpoint()  # Insert your handling here
     
            # Getting predicted values and log probs for given states and actions
            logprobs, state_values = self.policy.evaluate(states, actions)

            # NEW: Check for NaNs in state_values returned from the policy
            if self.debug:
                if torch.isnan(state_values).any() or torch.isinf(state_values).any():
                    print("Invalid values detected in 'state_values'")
                    breakpoint()  # Insert your handling here

            # Calculate the advantages
            advantages = returns - state_values.detach()
            # Normalize the advantages (optional, but can help in training stability)
            advantages = (advantages - advantages.mean()) / (advantages.std() + 1e-5)

            # NEW: Check for NaNs after normalizing advantages
            if self.debug:
                if torch.isnan(advantages).any() or torch.isinf(advantages).any():
                    print("Invalid values detected after normalizing 'advantages'")
                    breakpoint()  # Insert your handling here

            # Compute ratio for PPO
            old_logprobs, _ = self.policy_old.evaluate(states, actions)
            
            if self.debug:
                # Before calculating the ratio, add these checks
                if torch.isnan(logprobs).any() or torch.isinf(logprobs).any():
                    print("Anomaly detected in 'logprobs'")
                    breakpoint()  # Insert your handling here
                
                if torch.isnan(old_logprobs).any() or torch.isinf(old_logprobs).any():
                    print("Anomaly detected in 'old_logprobs'")
                    breakpoint()  # Insert your handling here

            log_diff = logprobs - old_logprobs.detach()
            clamp_value = 50
            log_diff_clamped = torch.clamp(log_diff, -clamp_value, clamp_value)  # clamp_value could be a large number like 50 or 100
            ratio = torch.exp(log_diff_clamped)

            if self.debug:
                # Checking critical tensors for NaNs
                if torch.isnan(ratio).any():
                    print("NaN detected in 'ratio'")
                    breakpoint()  # Insert your handling here
                
                if torch.isinf(ratio).any():
                    print("Inf detected in 'ratio'")
                    breakpoint()  # Insert your handling here
                
                if torch.isnan(advantages).any():
                    print("NaN detected in 'advantages'")
                    breakpoint()  # Insert your handling here

                if torch.isinf(advantages).any():
                    print("Inf detected in 'advantages'")
                    breakpoint()  # Insert your handling here

            # Surrogate loss
            surr1 = ratio * advantages
            surr2 = (
                torch.clamp(ratio, 1 - self.eps_clip, 1 + self.eps_clip) * advantages
            )
            policy_loss = -torch.min(surr1, surr2).mean()
            
            # Value loss
            value_loss = self.vl_coef * self.mse_loss(state_values, returns)

            # Entropy (for exploration)
            entropy_loss = -self.el_coef * logprobs.mean()

            if self.debug:
                if torch.isnan(policy_loss) or torch.isinf(policy_loss):
                    print("Anomaly in policy loss detected.")
                    # Additional checks can be placed here
                    self.check_gradients(self.policy)
                
                    breakpoint()  # Insert your handling here
                
                if torch.isnan(entropy_loss) or torch.isinf(entropy_loss):
                    print("Anomaly in entropy loss detected.")
                    breakpoint()  # Insert your handling here
    
            # Total loss
            loss = policy_loss + value_loss + entropy_loss
            
            if self.debug:
                if torch.isnan(loss):
                    print("NaN detected in loss")
                    # handle the issue as appropriate: stop training, log, raise exception, etc.
                    breakpoint()

            # Optimize policy network
            self.optimizer.zero_grad()
            loss.backward()

            # After backpropagation/Before optimizer step: Check if gradients are NaN
            self.check_gradients(self.policy.action_mu)

            self.policy.check_for_nan_gradients()

            # Clip gradients to prevent exploding gradients
            # Large gradients can cause the weights to update too aggressively. 
            # Actions can have extream values e.g.(100, -0.10, ...)
            torch.nn.utils.clip_grad_norm_(self.policy.parameters(), max_norm=0.7)
        
            self.optimizer.step()
            # After optimizer step: Check for NaN
            self.check_weights(self.policy.action_mu)

            if self.wandb_log:
                wandb.log(
                    {
                        "policy_loss": policy_loss.item(),
                        "value_loss": value_loss.item(),
                        "entropy_loss": entropy_loss.item(),
                        "total_loss": loss.item(),
                    }
                )
            
            if self.debug:
                # Check for NaNs in the model's parameters
                for name, param in self.policy.named_parameters():
                    if torch.isnan(param.data).any():
                        print(f"NaN detected in the model parameters: {name}")
                        breakpoint()

        # Copy new weights into old policy
        self.policy_old.load_state_dict(self.policy.state_dict())
        
        # Update learning rates using the lr_scheduler if provided
        self.iterations += 1
        if self.lr_scheduler is not None:
            self.lr_scheduler.step(self.optimizer, self.iterations)

        # Log the learning rates for each param_group if wandb_log is enabled
        if self.wandb_log:
            for i, param_group in enumerate(self.optimizer.param_groups):
                learning_rate = param_group['lr']
                wandb.log({f"learning_rate_group_{i}": learning_rate})

    def save(self, path):
        # Saving
        torch.save(
            {
                "model_state_dict": self.policy.state_dict(),
                "optimizer_state_dict": self.optimizer.state_dict(),
                "state_normalizer_state": self.state_normalizer.get_state(),
            },
            path,
        )

    def load(self, path):
        checkpoint = torch.load(path, map_location=self.device)
        self.policy.load_state_dict(checkpoint["model_state_dict"])
        self.policy_old.load_state_dict(self.policy.state_dict())
        self.optimizer.load_state_dict(checkpoint["optimizer_state_dict"])
        self.state_normalizer.set_state(checkpoint["state_normalizer_state"])



# filename: sinusoidal_positional_encoding.py

import torch
import torch.nn as nn
import math

class SinusoidalPositionalEncoding(nn.Module):
    def __init__(self, d_model, max_len=2000, device=torch.device("cpu")):
        super(SinusoidalPositionalEncoding, self).__init__()
        self.encoding = torch.zeros(max_len, d_model).to(device)
        position = torch.arange(0, max_len).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2).float() * -(math.log(10000.0) / d_model))
        div_term = torch.repeat_interleave(div_term, 2, dim=0)  # Repeat each element twice
        self.encoding[:, 0::2] = torch.sin(position * div_term[:d_model//2])
        self.encoding[:, 1::2] = torch.cos(position * div_term[:d_model//2])
        self.encoding = self.encoding.unsqueeze(0)

    def forward(self, x):
        return x + self.encoding[:, :x.size(1)].detach()



import torch
import torch.nn as nn
from torch.nn import MultiheadAttention
from nanoppo.sinusoidal_positional_encoding import SinusoidalPositionalEncoding

class ActorCriticCausalAttention(nn.Module):
    def __init__(self, state_dim, action_dim, nhead, action_low_tensor, action_high_tensor, device, rescale=False, debug=False):
        super(ActorCriticCausalAttention, self).__init__()
        self.action_low_tensor = action_low_tensor
        self.action_high_tensor = action_high_tensor
        self.rescale = rescale
        self.debug = debug
        self.epsilon = 1e-5
        self.positional_encoding = SinusoidalPositionalEncoding(d_model = state_dim, device=device)
        
        # Actor (Mu)
        self.action_mu = nn.Sequential(
            MultiheadAttention(state_dim, nhead, batch_first=True),
            nn.Linear(state_dim, nhead),
            #nn.Tanh(),
            nn.ReLU(),
            nn.Linear(nhead, action_dim)
        )
        
        # Actor (Log Std)
        self.action_log_std = nn.Sequential(
            MultiheadAttention(state_dim, nhead, batch_first=True),
            nn.Linear(state_dim, nhead),
            #nn.Tanh(),
            nn.ReLU(),
            nn.Linear(nhead, action_dim)
        )
        
        # Critic
        self.value_layer = nn.Sequential(
            MultiheadAttention(state_dim, nhead, batch_first=True),
            nn.Linear(state_dim, nhead),
            #nn.Tanh(),
            nn.ReLU(),
            nn.Linear(nhead, 1)
        )

    def forward(self, state): 
        # Validate state inputs at the very beginning of the method
        if self.debug:
            if torch.any(torch.isnan(state)):
                print("NaN detected in state input during forward pass.")
                breakpoint()  # Starts a pdb session so you can investigate.

        # Ensure state has at least 3 dimensions
        if len(state.shape) == 1:
            state = state.unsqueeze(0).unsqueeze(0)
        if len(state.shape) == 2:
            state = state.unsqueeze(0)
        # Create causal mask
        length = state.size(1)
        mask = torch.triu(torch.ones(length, length), diagonal=1).bool().to(state.device)
        
        state = self.positional_encoding(state)
        # Actor (Mu)
        attn_output_mu, _ = self.action_mu[0](state, state, state, attn_mask=mask)
        mu = self.action_mu[1:](attn_output_mu)

        # Validate outputs from the attention mechanism
        if self.debug:
            if torch.any(torch.isnan(attn_output_mu)):
                print("NaN detected in attn_output_mu during forward pass.")
                breakpoint()

        # Actor (Log Std)
        attn_output_std, _ = self.action_log_std[0](state, state, state, attn_mask=mask)

        # Validate outputs from the attention mechanism
        if self.debug:
            if torch.any(torch.isnan(attn_output_std)):
                print("NaN detected in attn_output_std during forward pass.")
                breakpoint()

        log_std = self.action_log_std[1:](attn_output_std)

        if self.debug:
            # Check for NaN values immediately after computation
            if torch.isnan(mu).any() or torch.isnan(log_std).any():
                print("NaN detected in 'mu' or 'log_std' during forward pass.")
                breakpoint() 

        std = torch.clamp(log_std.exp(), min=self.epsilon, max=1e2)

        # Validate the 'std' tensor after computation
        if self.debug:
            if torch.any(torch.isnan(std)):
                print("NaN detected in 'std' tensor during forward pass.")
                breakpoint()

        dist = torch.distributions.Normal(mu, std)
        return dist

    def act(self, state, action=None, compute_logprobs=True):
        # Ensure state has at least 3 dimensions
        if len(state.shape) == 1:
            state = state.unsqueeze(0).unsqueeze(0)
        if len(state.shape) == 2:
            state = state.unsqueeze(0)
        if self.debug:
            # Check if the state contains NaN values
            if torch.isnan(state).any():
                print("NaN detected in the state input of the 'act' method.")
                breakpoint()
        # Create causal mask for attention
        length = state.size(1)
        mask = torch.triu(torch.ones(length, length), diagonal=1).bool().to(state.device)

        state = self.positional_encoding(state)
        # Actor (Mu)
        attn_output_mu, _ = self.action_mu[0](state, state, state, attn_mask=mask)
        mu = self.action_mu[1:](attn_output_mu)
        mu = mu[:, -1, :]  # Take the last sequence element

        # Validate 'mu' after its computation
        if self.debug:
            if torch.any(torch.isnan(mu)):
                print("NaN detected in 'mu' tensor during act method.")
                breakpoint()
        
        # Actor (Log Std)
        attn_output_std, _ = self.action_log_std[0](state, state, state, attn_mask=mask)
        log_std = self.action_log_std[1:](attn_output_std)
        log_std = log_std[:, -1, :]  # Take the last sequence element

        # Validate 'log_std' after its computation
        if self.debug:
            if torch.any(torch.isnan(log_std)):
                print("NaN detected in 'log_std' tensor during act method.")
                breakpoint()

        std = torch.clamp(log_std.exp(), min=self.epsilon, max=1e2)
        if action is None:
            dist = torch.distributions.Normal(mu, std)
            action = dist.sample()
            action = torch.clamp(action, min=-10, max=10)
            # Rescale action to be in the desired range
            if self.rescale:
                action = self.rescale_action(action)

        if compute_logprobs:
            #dist = torch.distributions.Normal(mu, std)
            #logprobs = dist.log_prob(action).sum(axis=-1)
            # --- Start of modified code for calculating log probabilities ---
            variance = std.pow(2)
            # Probability density function calculation for normal distribution
            probs = torch.exp(-((action - mu).pow(2)) / (2 * variance)) / torch.sqrt(2 * torch.tensor(torch.pi).to(action.device) * variance)
        
            # Add a small epsilon to the probabilities to prevent taking log of zero
            safe_probs = probs + self.epsilon  # where self.epsilon is a small constant, e.g., 1e-10
        
            # Now, calculate log probabilities
            logprobs = torch.log(safe_probs).sum(axis=-1)  # You may not need to sum, depending on your specific use case

            if self.debug and (torch.isnan(logprobs).any() or torch.isinf(logprobs).any()):
                print("Anomaly detected in 'logprobs'")
                breakpoint()  # Stopping here if debugging is enabled

            # Sanitize logprobs to replace -inf with large negative numbers
            clean_logprobs = torch.where(
                torch.isinf(logprobs),
                torch.full_like(logprobs, -1e7),
                logprobs
            )
            return action, clean_logprobs
        return action

    def rescale_action(self, action):
        """Rescale action from [-1, 1] to [action_low, action_high]"""
        # It introduces training instability because of tanh's gradient
        # Add a small epsilon to the output of tanh to ensure that you don't get values exactly equal to -1 or 1
        # It helps to delays the inf policy loss or nan problem, but not completely solve it
        # TODO: Disable rescale for now
        action = torch.clamp(torch.tanh(action), -1 + self.epsilon, 1 - self.epsilon)

        # Adjust the range for the epsilon-clamped tanh
        adjusted_low = (-1 + self.epsilon)  # This becomes the new "minimum" of your tanh output
        adjusted_high = (1 - self.epsilon)  # This becomes the new "maximum" of your tanh output
    
        # This ensures that the output of tanh is linearly mapped from [adjusted_low, adjusted_high] to [action_low, action_high]
        action_range = self.action_high_tensor - self.action_low_tensor
        action = self.action_low_tensor + (action - adjusted_low) / (adjusted_high - adjusted_low) * action_range
        return action

    # Define the debug function
    def debug_this(self, model, state):
        print("Debugging Information:")
        # Check and print the weights and biases
        print("\nWeights and Biases in each layer:")
        for name, param in model.named_parameters():
            print(name, param.data)
        
        state = self.positional_encoding(state)
        # Pass the input through each layer individually and print the outputs
        x = state
        for i, layer in enumerate(model):
            x = layer(x)
            print(f"\nOutput after layer {i} ({layer.__class__.__name__}):")
            print(x)
            # If you find NaN at this stage, you can break the loop for efficiency
            if torch.isnan(x).any():
                print(f"Stopping early due to NaN at layer {i}")
                break

    def check_for_nan_gradients(self):
        if self.debug:
            for name, param in self.named_parameters():
                if param.grad is not None and torch.isnan(param.grad).any():
                    print(f"NaN detected in the gradients. Parameter: {name}")
                    # Here you can either call your debug function or raise an error
                    breakpoint()

    def get_value(self, state):
        # Ensure state has at least 3 dimensions
        if len(state.shape) == 1:
            state = state.unsqueeze(0).unsqueeze(0)
        if len(state.shape) == 2:
            state = state.unsqueeze(0)
        # Create causal mask for attention
        length = state.size(1)
        mask = torch.triu(torch.ones(length, length), diagonal=1).bool().to(state.device)
        
        state = self.positional_encoding(state)
        # Value Layer
        attn_output_value, _ = self.value_layer[0](state, state, state, attn_mask=mask)
        return self.value_layer[1:](attn_output_value)[:, -1, :]

    def evaluate(self, state, action):
        # Ensure state has at least 3 dimensions
        if len(state.shape) == 1:
            state = state.unsqueeze(0).unsqueeze(0)
        if len(state.shape) == 2:
            state = state.unsqueeze(0)
        # Create causal mask for attention
        length = state.size(1)
        mask = torch.triu(torch.ones(length, length), diagonal=1).bool().to(state.device)

        state = self.positional_encoding(state)
        # Actor (Mu)
        attn_output_mu, _ = self.action_mu[0](state, state, state, attn_mask=mask)
        mu = self.action_mu[1:](attn_output_mu)
        mu = mu[:, -1, :]  # Take the last sequence element
        if self.debug:
            # Check if 'mu' contains any NaN values and call the debug function if it does
            if torch.isnan(mu).any():
                print("NaN detected in output. Starting debug sequence...\n")
                self.debug_this(self.action_mu, state)
                breakpoint()

        # Actor (Log Std)
        attn_output_std, _ = self.action_log_std[0](state, state, state, attn_mask=mask)
        log_std = self.action_log_std[1:](attn_output_std)
        log_std = log_std[:, -1, :]  # Take the last sequence element

        std = log_std.exp()
        #dist = torch.distributions.Normal(mu, std)
        #logprobs = dist.log_prob(action).sum(axis=-1)
        variance = std.pow(2)
        # Probability density function calculation for normal distribution
        probs = torch.exp(-((action - mu).pow(2)) / (2 * variance)) / torch.sqrt(2 * torch.tensor(torch.pi).to(action.device) * variance)
    
        # Add a small epsilon to the probabilities to prevent taking log of zero
        safe_probs = probs + self.epsilon  # where self.epsilon is a small constant, e.g., 1e-10
    
        # Now, calculate log probabilities
        logprobs = torch.log(safe_probs).sum(axis=-1)  # You may not need to sum, depending on your specific use case

        if self.debug:
            # Before calculating the ratio, add these checks
            if (torch.isnan(logprobs).any() or torch.isinf(logprobs).any()):
                print("Anomaly detected in 'logprobs'")
                breakpoint()  # Insert your handling here
                
        # Sanitize logprobs to replace -inf with large negative numbers
        clean_logprobs = torch.where(
            torch.isinf(logprobs),
            torch.full_like(logprobs, -1e7),
            logprobs
        )

        # Value Layer
        attn_output_value, _ = self.value_layer[0](state, state, state, attn_mask=mask)
        state_value = self.value_layer[1:](attn_output_value)[:,-1,:]

        return clean_logprobs, torch.squeeze(state_value)




import math

class CosineLRScheduler:
    def __init__(self, warmup_iters, lr_decay_iters, learning_rate, min_lr):
        self.warmup_iters = warmup_iters
        self.lr_decay_iters = lr_decay_iters
        self.learning_rate = learning_rate
        self.min_lr = min_lr

    def set_optimizer(self, optimizer):
        self.optimizer = optimizer

    def get_lr_actor(self, it):
        if it < self.warmup_iters:
            return self.learning_rate * it / self.warmup_iters
        if it > self.lr_decay_iters:
            return self.min_lr
        decay_ratio = (it - self.warmup_iters) / (self.lr_decay_iters - self.warmup_iters)
        assert 0 <= decay_ratio <= 1
        coeff = 0.5 * (1.0 + math.cos(math.pi * decay_ratio))
        return self.min_lr + coeff * (self.learning_rate - self.min_lr)

    def get_lr_critic(self, it):
        # You can define a separate LR schedule for the critic if needed
        # Example: return self.learning_rate_critic * it / self.warmup_iters
        return self.get_lr_actor(it)

    def step(self, optimizer, iterations):
        for param_group in optimizer.param_groups:
            actor_lr = self.get_lr_actor(it=iterations)
            param_group['lr'] = actor_lr



import torch
import torch.nn as nn

class ActorCritic(nn.Module):
    def __init__(self, state_dim, action_dim, n_latent_var, action_low_tensor, action_high_tensor, rescale=False, debug=False):
        super(ActorCritic, self).__init__()
        self.action_low_tensor = action_low_tensor
        self.action_high_tensor = action_high_tensor
        self.rescale = rescale
        self.debug = False
        self.epsilon = 1e-5

        # Actor: outputs mean and log standard deviation
        self.action_mu = nn.Sequential(
            nn.Linear(state_dim, n_latent_var),
            nn.Tanh(),
            nn.Linear(n_latent_var, n_latent_var),
            nn.Tanh(),
            nn.Linear(n_latent_var, action_dim)
        )

        self.action_log_std = nn.Sequential(
            nn.Linear(state_dim, n_latent_var),
            nn.Tanh(),
            nn.Linear(n_latent_var, n_latent_var),
            nn.Tanh(),
            nn.Linear(n_latent_var, action_dim)
        )

        # Critic
        self.value_layer = nn.Sequential(
            nn.Linear(state_dim, n_latent_var),
            nn.Tanh(),
            nn.Linear(n_latent_var, n_latent_var),
            nn.Tanh(),
            nn.Linear(n_latent_var, 1)
        )

    def forward(self, state):
        mu = self.action_mu(state)
        log_std = self.action_log_std(state)

        if self.debug:
            # Check for NaN values immediately after computation
            if torch.isnan(mu).any() or torch.isnan(log_std).any():
                print("NaN detected in 'mu' or 'log_std' during forward pass.")
                breakpoint()
            
        # Avoid policy loss NAN or entropy Loss INF
        # std = log_std.exp() May cause NaN values in self.policy and Inf values in entropy_loss due to action_pro approaching zero.
        # The entropy of a policy in certain contexts is calculated using the probabilities of the actions, which the policy might take. 
        # If these probabilities approach zero, the log of near-zero probabilities can become negative infinity, 
        # and hence the entropy becomes positive infinity (since entropy is calculated as the expected value of -log(prob)). This is usually the cause of entropy_loss becoming inf.
        # Clamping the standard deviation to avoid extreme values
        std = torch.clamp(log_std.exp(), min=self.epsilon, max=1e2)  # you can adjust the max value based on your context
        
        dist = torch.distributions.Normal(mu, std)
        return dist

    def act(self, state, action=None, compute_logprobs=True):
        if self.debug:
            # Check if the state contains NaN values
            if torch.isnan(state).any():
                print("NaN detected in the state input of the 'act' method.")
                breakpoint()

        mu = self.action_mu(state)
        log_std = self.action_log_std(state)
        std = log_std.exp()
        dist = torch.distributions.Normal(mu, std)
        if action is None:
            action = dist.sample()
            # Rescale action to be in the desired range
            if self.rescale:
                action = self.rescale_action(action)

        if compute_logprobs:
            # policy_loss:NAN, entropy_loss:INF
            # Ensure that action values are within a valid range before computing log_prob
            # action = torch.clamp(action, self.action_low_tensor + self.epsilon, self.action_high_tensor - self.epsilon)
            logprobs = dist.log_prob(action).sum(axis=-1)

            if self.debug and (torch.isnan(logprobs).any() or torch.isinf(logprobs).any()):
                print("Anomaly detected in 'logprobs'")
                breakpoint()  # Stopping here if debugging is enabled
 
            # Sanitize logprobs to replace -inf with large negative numbers
            clean_logprobs = torch.where(
                torch.isinf(logprobs),
                torch.full_like(logprobs, -1e7),  # Replace -inf with -1e7
                logprobs
            )
            return action, clean_logprobs
        
        return action
    
    def rescale_action(self, action):
        """Rescale action from [-1, 1] to [action_low, action_high]"""
        # It introduces training instability because of tanh's gradient
        # Add a small epsilon to the output of tanh to ensure that you don't get values exactly equal to -1 or 1
        # It helps to delays the inf policy loss or nan problem, but not completely solve it
        # TODO: Disable rescale for now
        action = torch.clamp(torch.tanh(action), -1 + self.epsilon, 1 - self.epsilon)

        # Adjust the range for the epsilon-clamped tanh
        adjusted_low = (-1 + self.epsilon)  # This becomes the new "minimum" of your tanh output
        adjusted_high = (1 - self.epsilon)  # This becomes the new "maximum" of your tanh output
    
        # This ensures that the output of tanh is linearly mapped from [adjusted_low, adjusted_high] to [action_low, action_high]
        action_range = self.action_high_tensor - self.action_low_tensor
        action = self.action_low_tensor + (action - adjusted_low) / (adjusted_high - adjusted_low) * action_range
        return action

    # Define the debug function
    def debug_this(self, model):
        print("Debugging Information:")
        # Check and print the weights and biases
        print("\nWeights and Biases in each layer:")
        for name, param in model.named_parameters():
            print(name, param.data)
        
        # Pass the input through each layer individually and print the outputs
        x = state
        for i, layer in enumerate(model):
            x = layer(x)
            print(f"\nOutput after layer {i} ({layer.__class__.__name__}):")
            print(x)
            # If you find NaN at this stage, you can break the loop for efficiency
            if torch.isnan(x).any():
                print(f"Stopping early due to NaN at layer {i}")
                break

    def check_for_nan_gradients(self):
        if self.debug:
            for name, param in self.named_parameters():
                if param.grad is not None and torch.isnan(param.grad).any():
                    print(f"NaN detected in the gradients. Parameter: {name}")
                    # Here you can either call your debug function or raise an error
                    breakpoint()

    def evaluate(self, state, action):
        if self.rescale:
            assert (action >= self.action_low_tensor).all() and (action <= self.action_high_tensor).all(), "Actions are not rescaled!"

        mu = self.action_mu(state)
        if self.debug:
            # Check if 'mu' contains any NaN values and call the debug function if it does
            if torch.isnan(mu).any():
                print("NaN detected in output. Starting debug sequence...\n")
                self.debug_this(self.action_mu)
                breakpoint()
        log_std = self.action_log_std(state)
        std = log_std.exp()
        
        dist = torch.distributions.Normal(mu, std)
        logprobs = dist.log_prob(action).sum(axis=-1)

        if self.debug:
            # Before calculating the ratio, add these checks
            if (torch.isnan(logprobs).any() or torch.isinf(logprobs).any()):
                print("Anomaly detected in 'logprobs'")
                breakpoint()  # Insert your handling here

        # Sanitize logprobs to replace -inf with large negative numbers
        clean_logprobs = torch.where(
            torch.isinf(logprobs),
            torch.full_like(logprobs, -1e7),  # Replace -inf with -1e7
            logprobs
        )
        state_value = self.value_layer(state)
        return clean_logprobs, torch.squeeze(state_value)

    def get_value(self, state):
        return self.value_layer(state)
